#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <curses.h>

#include "pmacpci.h"

WINDOW *create_newwin(int w, int h, int x, int y);
void main_screen(WINDOW *wptr);
void connect_device(WINDOW *wptr);
void message(char *cptr);
void interactive();
void status_update();
void disconnect_device();
void plc_control();
void general_control();
void dpram_control();
void dpram_ioctl_read();
void download_plc();
void set_debug();
void plc_main_screen();
void dpram_main_screen();
void general_main_screen();
void enable_plc();
void disable_plc();
void clear_plc();
void list_plc();
void download_file();

extern int errno;
WINDOW *wm;
WINDOW *w1;
WINDOW *w2;
WINDOW *w3;
int fd = 0;
char msgs[3][60];
char device[40];

struct screen
{
  char title[40];
};



int main()
{
  char d;
  int mid;
  int nrows;
  int ncols;

  msgs[0][0] = '\0';
  msgs[1][0] = '\0';
  msgs[2][0] = '\0';
  strcpy(device, "None\0");

  wm = initscr();
  cbreak();
  noecho();
  curs_set(0);
  getmaxyx(wm, nrows, ncols);
  mid = (ncols / 2);
  clear();
  refresh();

  mvwprintw(wm, 1, (mid - 15), "PMAC Driver Test Program Suite");
  mvwprintw(wm, 2, (mid - 15), "==============================");

  mvwprintw(wm, 3, (mid - 12), "Observatory Sciences Ltd");
  mvwprintw(wm, 4, (mid - 12), "========================");


  mvwprintw(wm, 22, (mid - 15), "www.observatorysciences.co.uk");

  w1 = create_newwin((ncols - 35), 10, 5, 6);
  w2 = create_newwin((ncols - 10), 5, 5, 16);
  w3 = create_newwin(25, 10, (ncols - 30), 6);

  
  main_screen(w1);
  
  message("Ready...");
  status_update();

  while(1){
    d = getch();
    if (d == 'q') break;

    if (d == '1'){
      general_control();
      main_screen(w1);
    }

    if (d == '2'){
      dpram_control();
      main_screen(w1);
    }

    if (d == '3'){
      message("Interactive mode.  Type \\q to leave.");
      interactive();
      main_screen(w1);
    }

    if (d == '4'){
      plc_control();
      main_screen(w1);
    }
  }
  endwin();

  return 0;
}

void main_screen(WINDOW *wptr)
{
  int mid;
  int nrows;
  int ncols;
  getmaxyx(wptr, nrows, ncols);
  mid = (ncols / 2);

  wclear(wptr);
  box(wptr, 0, 0);

  noecho();
  curs_set(0);

  mvwprintw(wptr, 0, 4, "Main Menu");
  mvwprintw(wptr, 2, (mid - 15), "1 - General Options");
  mvwprintw(wptr, 3, (mid - 15), "2 - DPRAM Options");
  mvwprintw(wptr, 4, (mid - 15), "3 - Enter interactive ASCII mode");
  mvwprintw(wptr, 5, (mid - 15), "4 - PLC program control menu");
  mvwprintw(wptr, 7, (mid - 15), "q - Quit the program");

  wrefresh(wptr);
}


void connect_device(WINDOW *wptr)
{
  char sptr[256];
  char msg[80];
  int mid;
  int nrows;
  int ncols;
  getmaxyx(wptr, nrows, ncols);
  mid = (ncols / 2);

  wclear(wptr);
  box(wptr, 0, 0);
  mvwprintw(wptr, 2, (mid - 15), "Enter path for device: ");

  echo();
  curs_set(1);

  wrefresh(wptr);

  mvwgetstr(wptr, 4, (mid - 15), sptr);

  fd = open(sptr, O_RDWR);
  if (fd > 0){
    strcpy(msg, "Successfully opened ");
    strcat(msg, sptr);
    message(msg);
    strcpy(device, sptr);
  } else {
    strcpy(msg, "Error opening ");
    strcat(msg, sptr);
    message(msg);
    strcpy(device, "None\0");
  }

}

void disconnect_device()
{
  int result = 0;
  char msg[80];

  result = close(fd);
  if (result < 0){
    strcpy(msg, "Error closing ");
    strcat(msg, device);
    message(msg);
  } else {
    strcpy(msg, "Closed device ");
    strcat(msg, device);
    message(msg);
  }
  strcpy(device, "None\0");
  fd = 0;
}


void interactive()
{
  int cx = 3;
  int cy = 2;
  int cnt = 0;
  char sptr[80];
  char rptr[80];
  char txt[5][80];
  int ch;

  txt[0][0] = '\0';
  txt[1][0] = '\0';
  txt[2][0] = '\0';
  txt[3][0] = '\0';
  txt[4][0] = '\0';

  curs_set(1);
  wclear(w1);
  box(w1, 0, 0);
  mvwprintw(w1, 1, 1, "ASCII Communication interactive mode");

  noecho();
  /*
  curs_set(0);
  */

  wrefresh(w1);

  while(1){
    ch = mvwgetch(w1, 8, (cnt + 3));

    /*
    sprintf(sptr, "No: %d", ch);
    message(sptr);
    */
    
    if (ch == 127){
      cnt--;
      if (cnt < 0) cnt = 0;
      sptr[cnt] = '\0';
    } else if (ch == 10){
      sptr[cnt] = '\0';
      cnt = 0;

      if (!strcmp(sptr, "\\q")){
	break;
      }

      strcpy(txt[0], txt[1]);
      strcpy(txt[1], txt[2]);
      strcpy(txt[2], txt[3]);
      strcpy(txt[3], txt[4]);
      strcpy(txt[4], sptr);
      sptr[0] = '\0';

      /*
       * now send the string and see if there is a response
       */
      if (fd > 0){
	write(fd, txt[4], strlen(txt[4]));
	read(fd, rptr, 80);

	if (strlen(rptr) > 0){
	  strcpy(txt[0], txt[1]);
	  strcpy(txt[1], txt[2]);
	  strcpy(txt[2], txt[3]);
	  strcpy(txt[3], txt[4]);
	  strcpy(txt[4], rptr);
	}
      }

    } else {
      sptr[cnt] = ch;
      cnt++;
      sptr[cnt] = '\0';

    }
      wclear(w1);
      box(w1, 0, 0);
      mvwprintw(w1, 1, 1, "ASCII Communication interactive mode");
      mvwprintw(w1, 3, 3, txt[0]);
      mvwprintw(w1, 4, 3, txt[1]);
      mvwprintw(w1, 5, 3, txt[2]);
      mvwprintw(w1, 6, 3, txt[3]);
      mvwprintw(w1, 7, 3, txt[4]);
      mvwprintw(w1, 8, 3, sptr);
      wrefresh(w1);
  }

}

void message(char *cptr)
{
  strcpy(msgs[0], msgs[1]);
  strcpy(msgs[1], msgs[2]);
  strcpy(msgs[2], cptr);

  wclear(w2);
  box(w2, 0, 0);
  mvwprintw(w2, 0, 4, "Messages");

  mvwprintw(w2, 1, 1, msgs[0]);
  mvwprintw(w2, 2, 1, msgs[1]);
  mvwprintw(w2, 3, 1, msgs[2]);

  wrefresh(w2);
}

void status_update()
{
  char rptr[256];
  int dpram_start = 0;
  int dpram_finish = 0;
  int dpram_size = 0;
  wclear(w3);
  box(w3, 0, 0);
  mvwprintw(w3, 0, 4, "Status");
  mvwprintw(w3, 2, 2, "Device");
  mvwprintw(w3, 3, 3, device);
  mvwprintw(w3, 5, 2, "DPRAM Size");

  if (fd > 0){
    strcpy(rptr, "RX:$320F");
    write(fd, rptr, strlen(rptr));
    read(fd, rptr, 80);
    sscanf(rptr, "%d", &dpram_start);
    strcpy(rptr, "RX:$3210");
    write(fd, rptr, strlen(rptr));
    read(fd, rptr, 80);
    sscanf(rptr, "%d", &dpram_finish);
    dpram_size = (dpram_finish - dpram_start) * 4;
    sprintf(rptr, "%d bytes", dpram_size);
    mvwprintw(w3, 6, 3, rptr);
  }

  wrefresh(w3);
}

void plc_control()
{
  int d = 0;

  if (fd > 0){

    plc_main_screen();

    while(1){
      d = getch();
      if (d == 'm') break;

      if (d == '1'){
	download_plc();
	plc_main_screen();
      }

      if (d == '2'){
	list_plc();
	plc_main_screen();
      }

      if (d == '3'){
	enable_plc();
	plc_main_screen();
      }

      if (d == '4'){
	disable_plc();
	plc_main_screen();
      }

      if (d == '5'){
	clear_plc();
	plc_main_screen();
      }

    }
  } else {
    message("Connect to device first, then perform PLC operations.");
  }

}

void dpram_control()
{
  int d = 0;

  if (fd > 0){

    dpram_main_screen();

    while(1){
      d = getch();
      if (d == 'm') break;

      if (d == '1'){
	dpram_ioctl_read();
	dpram_main_screen();
      }

      if (d == '2'){
	list_plc();
	plc_main_screen();
      }

      if (d == '3'){
	enable_plc();
	plc_main_screen();
      }

      if (d == '4'){
	disable_plc();
	plc_main_screen();
      }

      if (d == '5'){
	clear_plc();
	plc_main_screen();
      }

    }
  } else {
    message("Connect to device first, then perform DPRAM operations.");
  }

}

void dpram_main_screen()
{
  int mid;
  int nrows;
  int ncols;
  int d = 0;
  getmaxyx(w1, nrows, ncols);
  mid = (ncols / 2);

  wclear(w1);
  box(w1, 0, 0);
  noecho();
  curs_set(0);
  mvwprintw(w1, 0, 4, "DPRAM Options");
  mvwprintw(w1, 2, (mid - 15), "1 - Read bytes using ioctl");
  /*
  mvwprintw(w1, 3, (mid - 15), "2 - Disconnect from device");
  mvwprintw(w1, 4, (mid - 15), "3 - Set Debug Level");
  mvwprintw(w1, 5, (mid - 15), "4 - Download pmc or instruction set");
  */
  mvwprintw(w1, 7, (mid - 15), "m - Return to the main menu");

  wrefresh(w1);
}

void dpram_ioctl_read()
{
  int d = 0;
  int j = 0;
  int mid;
  int nrows;
  int ncols;
  int nbytes;
  int location;
  struct _Device_Ext dev;

  if (fd > 0){
    getmaxyx(w1, nrows, ncols);
    mid = (ncols / 2);
    echo();
    curs_set(1);
    wclear(w1);
    box(w1, 0, 0);
    mvwprintw(w1, 2, (mid - 15), "Enter the number of bytes to read");
    wrefresh(w1); 
    mvwscanw(w1, 4, (mid - 15), "%d", &nbytes);

    wclear(w1);
    box(w1, 0, 0);
    mvwprintw(w1, 2, (mid - 15), "Enter the memory location to read");
    wrefresh(w1); 
    mvwscanw(w1, 4, (mid - 15), "%x", &location);

    dev.dpr_offset = ((location - 0x060000) * 4);
    dev.dpr_count = 4;

    ioctl(fd, 1, &dev);

    wclear(w1);
    box(w1, 0, 0);
    mvwprintw(w1, 2, (mid - 15), "Memory location: %x", location);
    for (j = 0; j < nbytes; j++){
      mvwprintw(w1, 4, (mid - (15 - (j*4))), "%d", dev.buffer[j]);
    }

    wrefresh(w1); 

    noecho();
    curs_set(0);
    while(1){
      d = getch();
      if (d == 'q') break;
    }

  } else {
    message("Connect to a device before attempting to read memory.");
  }
}


void general_control()
{
  int d = 0;
  general_main_screen();

  while(1){
    d = getch();
    if (d == 'm') break;
    
    if (d == '1'){
      connect_device(w1);
      status_update();
      general_main_screen();
    }

    if (d == '2'){
      disconnect_device();
      status_update();
      general_main_screen();
    }
    
    if (d == '3'){
      set_debug();
      general_main_screen();
    }
      
    if (d == '4'){
      download_file();
      general_main_screen();
    }
  }
}

void general_main_screen()
{
  int mid;
  int nrows;
  int ncols;
  int d = 0;
  getmaxyx(w1, nrows, ncols);
  mid = (ncols / 2);

  wclear(w1);
  box(w1, 0, 0);
  noecho();
  curs_set(0);
  mvwprintw(w1, 0, 4, "General Options");
  mvwprintw(w1, 2, (mid - 15), "1 - Connect to device");
  mvwprintw(w1, 3, (mid - 15), "2 - Disconnect from device");
  mvwprintw(w1, 4, (mid - 15), "3 - Set Debug Level");
  mvwprintw(w1, 5, (mid - 15), "4 - Download pmc or instruction set");
  /*
  mvwprintw(w1, 5, (mid - 15), "4 - Disable PLC program");
  mvwprintw(w1, 6, (mid - 15), "5 - Clear PLC program");
  */
  mvwprintw(w1, 7, (mid - 15), "m - Return to the main menu");

  wrefresh(w1);
}

void set_debug()
{
  int mid;
  int nrows;
  int ncols;
  int debug;

  if (fd > 0){
    getmaxyx(w1, nrows, ncols);
    mid = (ncols / 2);
    echo();
    curs_set(1);
    wclear(w1);
    box(w1, 0, 0);
    mvwprintw(w1, 2, (mid - 15), "Enter new debug level (0-5):");
    wrefresh(w1); 
    mvwscanw(w1, 4, (mid - 15), "%d", &debug);
    ioctl(fd, 6, debug);
  } else {
    message("Connect to a device before setting the debug");
  }
}

void plc_main_screen()
{
  int mid;
  int nrows;
  int ncols;
  int d = 0;
  getmaxyx(w1, nrows, ncols);
  mid = (ncols / 2);

  wclear(w1);
  box(w1, 0, 0);
  noecho();
  curs_set(0);
  mvwprintw(w1, 0, 4, "PLC Control");
  mvwprintw(w1, 2, (mid - 15), "1 - Download PLC program");
  mvwprintw(w1, 3, (mid - 15), "2 - List PLC program");
  mvwprintw(w1, 4, (mid - 15), "3 - Enable PLC program");
  mvwprintw(w1, 5, (mid - 15), "4 - Disable PLC program");
  mvwprintw(w1, 6, (mid - 15), "5 - Clear PLC program");

  mvwprintw(w1, 7, (mid - 15), "m - Return to the main menu");

  wrefresh(w1);
}

void download_plc()
{
  char sptr[256];
  char rptr[256];
  char msg[80];
  int mid;
  int nrows;
  int ncols;
  int plc_no;
  FILE *plc_file;
  getmaxyx(w1, nrows, ncols);
  mid = (ncols / 2);

  wclear(w1);
  box(w1, 0, 0);
  mvwprintw(w1, 2, (mid - 15), "Enter filename (inc path) of");
  mvwprintw(w1, 3, (mid - 15), "plc program:");

  echo();
  curs_set(1);

  wrefresh(w1);

  mvwgetstr(w1, 5, (mid - 15), sptr);

  wclear(w1);
  box(w1, 0, 0);
  mvwprintw(w1, 2, (mid - 15), "Enter PLC number (0-31):");
  wrefresh(w1);

  mvwscanw(w1, 4, (mid - 15), "%d", &plc_no);

  plc_file = fopen(sptr, "r");

  /*
   * Assuming all above went ok now open the file, read it in and
   * download it to the pmac card.
   */
  if (plc_file == NULL){
    sprintf(rptr, "Could not open %s, PLC not downloaded", sptr);
    message(rptr);
  } else {
    write(fd, "CLOSE", 5);
    read(fd, rptr, 80);
    write(fd, "DELETE GATHER", 13);
    read(fd, rptr, 80);
    sprintf(rptr, "OPEN PLC %d", plc_no);
    write(fd, rptr, strlen(rptr));
    read(fd, rptr, 80);
    write(fd, "CLEAR", 5);
    read(fd, rptr, 80);
    while(fgets(rptr, 256, plc_file) != NULL){
      write(fd, rptr, strlen(rptr));
      read(fd, rptr, 80);
    }
    write(fd, "CLOSE", 5);
    read(fd, rptr, 80);
  
    fclose(plc_file);
  }

  noecho();
  curs_set(0);
}

void enable_plc()
{
  int mid;
  int nrows;
  int ncols;
  int plc_no;
  char rptr[128];

  getmaxyx(w1, nrows, ncols);
  mid = (ncols / 2);

  wclear(w1);
  box(w1, 0, 0);

  echo();
  curs_set(1);

  mvwprintw(w1, 2, (mid - 15), "Enter PLC number (0-31) to enable:");
  mvwprintw(w1, 3, (mid - 15), "-1 to abort.");
  wrefresh(w1);

  mvwscanw(w1, 5, (mid - 15), "%d", &plc_no);

  if (plc_no > -1 && plc_no < 32){
    sprintf(rptr, "ENABLE PLC %d", plc_no);
    write(fd, rptr, strlen(rptr));
    read(fd, rptr, 80);
  }
}

void disable_plc()
{
  int mid;
  int nrows;
  int ncols;
  int plc_no;
  char rptr[128];

  getmaxyx(w1, nrows, ncols);
  mid = (ncols / 2);

  wclear(w1);
  box(w1, 0, 0);

  echo();
  curs_set(1);

  mvwprintw(w1, 2, (mid - 15), "Enter PLC number (0-31) to disable:");
  mvwprintw(w1, 3, (mid - 15), "-1 to abort.");
  wrefresh(w1);

  mvwscanw(w1, 5, (mid - 15), "%d", &plc_no);

  if (plc_no > -1 && plc_no < 32){
    sprintf(rptr, "DISABLE PLC %d", plc_no);
    write(fd, rptr, strlen(rptr));
    read(fd, rptr, 80);
  }
}

void clear_plc()
{
  int mid;
  int nrows;
  int ncols;
  int plc_no;
  char rptr[128];

  getmaxyx(w1, nrows, ncols);
  mid = (ncols / 2);

  wclear(w1);
  box(w1, 0, 0);

  echo();
  curs_set(1);

  mvwprintw(w1, 2, (mid - 15), "Enter PLC number (0-31) to clear:");
  mvwprintw(w1, 3, (mid - 15), "-1 to abort.");
  wrefresh(w1);

  mvwscanw(w1, 5, (mid - 15), "%d", &plc_no);

  if (plc_no > -1 && plc_no < 32){
    sprintf(rptr, "OPEN PLC %d", plc_no);
    write(fd, rptr, strlen(rptr));
    read(fd, rptr, 80);
    write(fd, "CLEAR", 5);
    read(fd, rptr, 80);
  }
}

void list_plc()
{
  int mid;
  int nrows;
  int ncols;
  int plc_no;
  char rptr[128];
  char buf[128][128];
  int max = 0;
  int pos = 0;
  int j = 0;
  int d = 0;

  for (j = 0; j < 128; j++){
    buf[j][0] = '\0';
  }

  getmaxyx(w1, nrows, ncols);
  mid = (ncols / 2);

  wclear(w1);
  box(w1, 0, 0);

  echo();
  curs_set(1);

  mvwprintw(w1, 2, (mid - 15), "Enter PLC number (0-31) to list:");
  mvwprintw(w1, 3, (mid - 15), "-1 to abort.");
  wrefresh(w1);

  mvwscanw(w1, 5, (mid - 15), "%d", &plc_no);

  if (plc_no > -1 && plc_no < 32){
    sprintf(rptr, "LIST PLC %d", plc_no);
    write(fd, rptr, strlen(rptr));
    while(strlen(rptr) > 0){
      read(fd, rptr, 80);
      strcpy(buf[max], rptr);
      max++;
    }

    keypad(stdscr, TRUE);
    noecho();
    curs_set(0);
    while(1){
      wclear(w1);
      box(w1, 0, 0);
      for (j = pos; j < (pos + 5); j++){
	mvwprintw(w1, (j - pos + 2), (mid - 19), "%d", (j+1));
	mvwprintw(w1, (j - pos + 2), (mid - 15), buf[j]);
      }
      mvwprintw(w1, 8, (mid - 19), "Use arrow keys, q to exit");
      wrefresh(w1);

      d = getch();
      if (d == 'q') break;

      if (d == KEY_DOWN){
	pos++;
	if ((pos + 6) > max){
	  pos = max - 6;
	  if (pos < 0) pos = 0;
	}
      }

      if (d == KEY_UP){
	pos--;
	if (pos < 0) pos = 0;
      }
    }
  }
}

void download_file()
{
  char sptr[256];
  char rptr[256];
  char *wstr;
  char msg[80];
  int mid;
  int nrows;
  int ncols;
  int plc_no;
  int wt = 0;
  int line = 0;
  FILE *plc_file;
  getmaxyx(w1, nrows, ncols);
  mid = (ncols / 2);

  wclear(w1);
  box(w1, 0, 0);
  mvwprintw(w1, 2, (mid - 15), "Enter filename (inc path) of");
  mvwprintw(w1, 3, (mid - 15), "plc program:");

  echo();
  curs_set(1);

  wrefresh(w1);

  mvwgetstr(w1, 5, (mid - 15), sptr);

  wclear(w1);
  box(w1, 0, 0);

  plc_file = fopen(sptr, "r");

  /*
   * Assuming all above went ok now open the file, read it in and
   * download it to the pmac card.
   */
  if (plc_file == NULL){
    sprintf(rptr, "Could not open %s, file not downloaded", sptr);
    message(rptr);
  } else {
    while(fgets(rptr, 256, plc_file) != NULL){
      line++;
      wclear(w1);
      box(w1, 0, 0);
      sprintf(msg, "Downloading line %d", line, rptr);
      mvwprintw(w1, 2, (mid - 15), msg);
      wrefresh(w1);
      if (!strncmp("wait", rptr, 4)){
	wstr = rptr+5;
	sscanf(wstr, "%d", &wt);
	sleep(wt);
      } else {
	write(fd, rptr, strlen(rptr));
	read(fd, rptr, 80);
      }
    }  
    fclose(plc_file);
  }

  noecho();
  curs_set(0);
}

WINDOW *create_newwin(int w, int h , int x, int y)
{
  WINDOW *wn;

  wn = newwin(h, w, y, x);
  box(wn, 0, 0);
  wrefresh(wn);

  return wn;
}
